#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from .mib_db_api import *
from voltha_protos.omci_alarm_db_pb2 import AlarmInstanceData, AlarmClassData, \
    AlarmDeviceData, AlarmAttributeData
from pyvoltha.common.config.config_backend import EtcdStore
from pyvoltha.common.utils.registry import registry
import six
from six.moves import range

class AlarmDbExternal(MibDbApi):
    """
    A persistent external OpenOMCI Alarm Database
    """
    CURRENT_VERSION = 1                       # VOLTHA v1.3.0 release
    ALARM_BITMAP_KEY = 'alarm_bit_map'

    _TIME_FORMAT = '%Y%m%d-%H%M%S.%f'

    # Paths from root proxy
    ALARM_PATH = '/omci_alarms'
    DEVICE_PATH = '{}'            # .format(device_id)

    # Classes, Instances, and Attributes as lists from root proxy
    CLASS_PATH = DEVICE_PATH + '/classes/{}'                                # .format(device_id, class_id)
    INSTANCE_PATH = DEVICE_PATH + '/classes/{}/instances/{}'                 # .format(device_id, class_id, instance_id)

    def __init__(self, omci_agent):
        """
        Class initializer
        :param omci_agent: (OpenOMCIAgent) OpenOMCI Agent
        """
        super(AlarmDbExternal, self).__init__(omci_agent)

        self.args = registry('main').get_args()
        host, port = self.args.etcd.split(':', 1)
        self._kv_store = EtcdStore(host, port, AlarmDbExternal.ALARM_PATH)

    def start(self):
        """
        Start up/restore the database
        """
        self.log.debug('start')

        if not self._started:
            super(AlarmDbExternal, self).start()

    def stop(self):
        """
        Start up the database
        """
        self.log.debug('stop')

        if self._started:
            super(AlarmDbExternal, self).stop()
            # TODO: Delete this method if nothing else is done except calling the base class

    def _time_to_string(self, time):
        return time.strftime(AlarmDbExternal._TIME_FORMAT) if time is not None else ''

    def _string_to_time(self, time):
        return datetime.strptime(time, AlarmDbExternal._TIME_FORMAT) if len(time) else None

    def _attribute_to_string(self, value):
        """
        Convert an ME's attribute value to string representation

        :param value: (long) Alarm bitmaps are always a Long
        :return: (str) String representation of the value
        """
        return str(value)

    def _string_to_attribute(self, str_value):
        """
        Convert an ME's attribute value-string to its Scapy decode equivalent

        :param device_id: (str) ONU Device ID
        :param class_id: (int) Class ID
        :param attr_name: (str) Attribute Name (see EntityClasses)
        :param str_value: (str) Attribute Value in string form

        :return: (various) String representation of the value
        :raises KeyError: Device, Class ID, or Attribute does not exist
        """
        # Alarms are always a bitmap which is a long
        return int(str_value) if len(str_value) else 0

    def add(self, device_id, overwrite=False):
        """
        Add a new ONU to database

        :param device_id: (str) Device ID of ONU to add
        :param overwrite: (bool) Overwrite existing entry if found.

        :raises KeyError: If device already exists and 'overwrite' is False
        """
        self.log.debug('add-device', device_id=device_id, overwrite=overwrite)

        now = datetime.utcnow()
        found = False

        data = AlarmDeviceData(device_id=device_id,
                               created=self._time_to_string(now),
                               version=AlarmDbExternal.CURRENT_VERSION,
                               last_alarm_sequence=0)
        path = self._get_device_path(device_id)
        self.log.debug("add-device-path", device_id=device_id, path=path)
        try:
            #raises KeyError if not found
            device = self._kv_store[path]

            #device is found at this point
            found = True

            if not overwrite:
                # Device already exists
                raise KeyError('Device with ID {} already exists in Alarm database'.
                               format(device_id))

            # Overwrite with new data
            self._kv_store[path] = data.SerializeToString()
            self._modified = now

        except KeyError:
            if found:
                raise
            # Did not exist, add it now
            self._kv_store[path] = data.SerializeToString()
            self._created = now
            self._modified = now

    def remove(self, device_id):
        """
        Remove an ONU from the database

        :param device_id: (str) Device ID of ONU to remove from database
        """
        self.log.debug('remove-device', device_id=device_id)

        if not self._started:
            raise DatabaseStateError('The Database is not currently active')

        if not isinstance(device_id, six.string_types):
            raise TypeError('Device ID should be an string')

        try:
            path = self._get_device_path(device_id)
            del self._kv_store[path]
            self._modified = datetime.utcnow()

        except KeyError:
            # Did not exists, which is not a failure
            pass

        except Exception as e:
            self.log.exception('remove-exception', device_id=device_id, e=e)
            raise

    def _get_device_path(self, device_id):
        return AlarmDbExternal.DEVICE_PATH.format(device_id)

    def _get_class_path(self, device_id, class_id):
        if not self._started:
            raise DatabaseStateError('The Database is not currently active')

        if not 0 <= class_id <= 0xFFFF:
            raise ValueError('class-id is 0..0xFFFF')

        fmt = AlarmDbExternal.CLASS_PATH
        return fmt.format(device_id, class_id)


    def _get_instance_path(self, device_id, class_id, instance_id):
        if not self._started:
            raise DatabaseStateError('The Database is not currently active')

        if not isinstance(device_id, six.string_types):
            raise TypeError('Device ID is a string')

        if not 0 <= class_id <= 0xFFFF:
            raise ValueError('class-id is 0..0xFFFF')

        if not 0 <= instance_id <= 0xFFFF:
            raise ValueError('instance-id is 0..0xFFFF')

        fmt = AlarmDbExternal.INSTANCE_PATH
        return fmt.format(device_id, class_id, instance_id)

    def save_last_sync_time(self, device_id, value):
        """
        Save the Last Sync time to the database in an easy location to access

        :param device_id: (str) ONU Device ID
        :param value: (DateTime) Value to save
        """
        self.log.debug('save-last-sync-time', device_id=device_id, time=str(value))

        try:
            if not isinstance(value, datetime):
                raise TypeError('Expected a datetime object, got {}'.
                                format(type(datetime)))

            data = AlarmDeviceData()
            path = self._get_device_path(device_id)
            data.ParseFromString(self._kv_store[path])

            now = datetime.utcnow()
            data.last_sync_time = self._time_to_string(value)

            # Update
            self._kv_store[path] = data.SerializeToString()
            self._modified = now
            self.log.debug('save-sync-time-complete', device_id=device_id)

        except Exception as e:
            self.log.exception('save-last-sync-exception', device_id=device_id, e=e)
            raise

    def get_last_sync_time(self, device_id):
        """
        Get the Last Sync Time saved to the database for a device

        :param device_id: (str) ONU Device ID
        :return: (int) The Value or None if not found
        """
        self.log.debug('get-last-sync-time', device_id=device_id)

        try:
            data = AlarmDeviceData()
            path = self._get_device_path(device_id)
            data.ParseFromString(self._kv_store[path])
            return self._string_to_time(data.last_sync_time)

        except KeyError:
            return None     # OMCI MIB_DB entry has not yet been created

        except Exception as e:
            self.log.exception('get-last-sync-time-exception', e=e)
            raise

    def save_alarm_last_sync(self, device_id, value):
        """
        Save the Last Alarm Sequence value to the database in an easy location to access

        :param device_id: (str) ONU Device ID
        :param value: (int) Value to save
        """
        self.log.debug('save-last-sync', device_id=device_id, seq=str(value))

        try:
            if not isinstance(value, int):
                raise TypeError('Expected a integer, got {}'.format(type(value)))

            data = AlarmDeviceData()
            path = self._get_device_path(device_id)
            data.ParseFromString(self._kv_store[path])

            now = datetime.utcnow()
            data.last_alarm_sequence = int(value)

            # Update
            self._kv_store[path] = data.SerializeToString()
            self._modified = now
            self.log.debug('save-sequence-complete', device_id=device_id)

        except Exception as e:
            self.log.exception('save-last-sync-exception', device_id=device_id, e=e)
            raise

    def get_alarm_last_sync(self, device_id):
        """
        Get the Last Sync Time saved to the database for a device

        :param device_id: (str) ONU Device ID
        :return: (int) The Value or None if not found
        """
        self.log.debug('get-last-sync', device_id=device_id)

        try:
            data = AlarmDeviceData()
            path = self._get_device_path(device_id)
            data.ParseFromString(self.kv_store[path])
            return int(data.last_alarm_sequence)

        except KeyError:
            return None     # OMCI ALARM_DB entry has not yet been created

        except Exception as e:
            self.log.exception('get-last-alarm-exception', e=e)
            raise

    def _add_new_class_and_instance(self, device_id, class_id, instance_id, attributes):
        """
        Create an entry for a new class in the external database

        :param device_id: (str) ONU Device ID
        :param class_id: (int) ME Class ID
        :param instance_id: (int) ME Entity ID
        :param attributes: (dict) Attribute dictionary

        :returns: (bool) True if the value was saved to the database. False if the
                         value was identical to the current instance
        """
        self.log.debug('add-new-class-and-instance', device_id=device_id, class_id=class_id,
                       instance_id=instance_id, attributes=attributes)

        class_data = AlarmClassData(class_id=class_id)

        class_path = AlarmDbExternal.CLASS_PATH.format(device_id, class_id)
        self._kv_store[class_path] = class_data.SerializeToString()

        inst_data = self._create_new_instance(device_id, class_id, instance_id, attributes)
        inst_path = AlarmDbExternal.INSTANCE_PATH.format(device_id, class_id, instance_id)
        self._kv_store[inst_path] = inst_data.SerializeToString()

        self.log.debug('set-complete', device_id=device_id, class_id=class_id,
                       entity_id=instance_id, attributes=attributes)
        return True

    def _create_new_instance(self,  device_id, class_id, instance_id, attributes):
        """
        Create an entry for a instance of an existing class in the external database

        :param device_id: (str) ONU Device ID
        :param class_id: (int) ME Class ID
        :param instance_id: (int) ME Entity ID
        :param attributes: (dict) Attribute dictionary

        :returns: (AlarmInstanceData) The new instance object
        """
        self.log.debug('create-new-instance', device_id=device_id, class_id=class_id,
                       instance_id=instance_id, attributes=attributes)

        now = self._time_to_string(datetime.utcnow())
        attrs = [AlarmAttributeData(name=k,
                                    value=self._attribute_to_string(v)) for k, v in attributes.items()]
        instance_data = AlarmInstanceData(instance_id=instance_id,
                                          created=now,
                                          modified=now,
                                          attributes=attrs)

        self.log.debug('create-new-instance-complete', device_id=device_id, class_id=class_id,
                       entity_id=instance_id, attributes=attributes)
        return instance_data

    def set(self, device_id, class_id, instance_id, attributes):
        """
        Set a database value.  This should only be called by the Alarm synchronizer
        and its related tasks

        :param device_id: (str) ONU Device ID
        :param class_id: (int) ME Class ID
        :param instance_id: (int) ME Entity ID
        :param attributes: (dict) Attribute dictionary

        :returns: (bool) True if the value was saved to the database. False if the
                         value was identical to the current instance

        :raises KeyError: If device does not exist
        :raises DatabaseStateError: If the database is not enabled
        """
        self.log.debug('set', device_id=device_id, class_id=class_id,
                       instance_id=instance_id, attributes=attributes)
        try:
            if not isinstance(device_id, six.string_types):
                raise TypeError('Device ID should be a string')

            if not 0 <= class_id <= 0xFFFF:
                raise ValueError("Invalid Class ID: {}, should be 0..65535".format(class_id))

            if not 0 <= instance_id <= 0xFFFF:
                raise ValueError("Invalid Instance ID: {}, should be 0..65535".format(instance_id))

            if not isinstance(attributes, dict):
                raise TypeError("Attributes should be a dictionary")

            if not self._started:
                raise DatabaseStateError('The Database is not currently active')

            try:
                class_data = AlarmClassData()
                class_path = AlarmDbExternal.CLASS_PATH.format(device_id, class_id)
                class_data.ParseFromString(self._kv_store[class_path])

                modified = False
                new_data = None
                try:
                    inst_data = AlarmInstanceData()
                    inst_path = AlarmDbExternal.INSTANCE_PATH.format(device_id, class_id, instance_id)
                    inst_data.ParseFromString(self._kv_store[inst_path])

                    exist_attr_indexes = dict()
                    attr_len = len(inst_data.attributes)

                    for index in range(0, attr_len):
                        exist_attr_indexes[inst_data.attributes[index].name] = index

                    str_value = ''
                    new_attributes = []

                    for k, v in attributes.items():
                        try:
                            str_value = self._attribute_to_string(v)
                            new_attributes.append(AlarmAttributeData(name=k, value=str_value))

                        except Exception as e:
                            self.log.exception('save-error', e=e, class_id=class_id,
                                            attr=k, value_type=type(v))

                        if k not in exist_attr_indexes or \
                                inst_data.attributes[exist_attr_indexes[k]].value != str_value:
                            modified = True
                            self.log.debug('different-attributes', device_id=device_id, class_id=class_id,
                                    instance_id=instance_id, attributes=inst_data.attributes, new_attributes=new_attributes)

                    if modified:
                        now = datetime.utcnow()
                        new_data = AlarmInstanceData(instance_id=instance_id,
                                                    created=inst_data.created,
                                                    modified=self._time_to_string(now),
                                                    attributes=new_attributes)

                except KeyError:
                    # Here if the instance_id does not yet exist in the database
                    new_data = self._create_new_instance(device_id, class_id, instance_id, attributes)
                    modified = True

                if modified:
                    del self._kv_store[inst_path]
                    self._kv_store[inst_path] = new_data.SerializeToString()
                return modified

            except KeyError:
                # Here if the class-id does not yet exist in the database
                return self._add_new_class_and_instance(device_id, class_id, instance_id,
                                           attributes)
        except Exception as e:
            self.log.exception('set-exception', device_id=device_id, class_id=class_id,
                               instance_id=instance_id, attributes=attributes, e=e)
            raise

    def delete(self, device_id, class_id, entity_id):
        """
        Delete an entity from the database if it exists.  If all instances
        of a class are deleted, the class is deleted as well.

        :param device_id: (str) ONU Device ID
        :param class_id: (int) ME Class ID
        :param entity_id: (int) ME Entity ID

        :returns: (bool) True if the instance was found and deleted. False
                         if it did not exist.

        :raises KeyError: If device does not exist
        :raises DatabaseStateError: If the database is not enabled
        """
        self.log.debug('delete', device_id=device_id, class_id=class_id,
                       entity_id=entity_id)

        if not self._started:
            raise DatabaseStateError('The Database is not currently active')

        if not isinstance(device_id, six.string_types):
            raise TypeError('Device ID should be an string')

        if not 0 <= class_id <= 0xFFFF:
            raise ValueError('class-id is 0..0xFFFF')

        if not 0 <= entity_id <= 0xFFFF:
            raise ValueError('instance-id is 0..0xFFFF')

        try:
            instpath = self._get_instance_path(device_id, class_id, entity_id)
            del self._kv_store[instpath]

            self._modified = datetime.utcnow()
            return True

        except KeyError:
            return False    # Not found

        except Exception as e:
            self.log.exception('get-last-data-exception', device_id=device_id, e=e)
            raise

    def query(self, device_id, class_id=None, instance_id=None, attributes=None):
        """
        Get database information.

        This method can be used to request information from the database to the detailed
        level requested

        :param device_id: (str) ONU Device ID
        :param class_id:  (int) Managed Entity class ID
        :param instance_id: (int) Managed Entity instance
        :param attributes: (list/set or str) Managed Entity instance's attributes

        :return: (dict) The value(s) requested. If class/inst/attribute is
                        not found, an empty dictionary is returned
        :raises KeyError: If the requested device does not exist
        :raises DatabaseStateError: If the database is not enabled
        """
        self.log.debug('query', device_id=device_id, class_id=class_id,
                       entity_instance_id=instance_id, attributes=attributes)
        data = dict()

        try:
            if class_id is None:
                # Get full device info
                self.log.warn('get-all-instances-by-device-not-supported')
                try:
                    dev_data = AlarmDeviceData()
                    device_path = self._get_device_path(device_id)
                    dev_data.ParseFromString(self._kv_store[device_path])
                    data = self._device_to_dict(dev_data)

                except KeyError:
                    data = dict()

            elif instance_id is None:
                # Get all instances of the class
                self.log.warn('get-all-instances-by-class-not-supported')
                try:
                    cls_data = AlarmClassData()
                    class_path = self._get_class_path(device_id, class_id)
                    cls_data.ParseFromString(self._kv_store[class_path])
                    data = self._class_to_dict(cls_data)

                except KeyError:
                    data = dict()

            else:
                # Get all attributes of a specific ME
                try:
                    inst_data = AlarmInstanceData()
                    inst_path = self._get_instance_path(device_id, class_id, instance_id)
                    inst_data.ParseFromString(self._kv_store[inst_path])

                    if attributes is None:
                        # All Attributes
                        data = self._instance_to_dict(inst_data)
                        self.log.debug('query-result-all', data=data)
                    else:
                        # Specific attribute(s)
                        if isinstance(attributes, six.string_types):
                            attributes = {attributes}

                        data = {
                            attr.name: self._string_to_attribute(attr.value)
                            for attr in inst_data.attributes if attr.name in attributes}

                except KeyError:
                    self.log.debug('no-instance-data-keyError')
                    data = dict()

                except Exception as e:
                    self.log.exception('get-last-sync-exception', device_id=device_id, e=e)
                    raise

            self.log.debug('query-result', data=data)
            return data

        except KeyError:
            self.log.warn('query-no-device', device_id=device_id)
            raise

        except Exception as e:
            self.log.exception('get-last-sync-exception', device_id=device_id, e=e)
            raise

    def _instance_to_dict(self, instance):
        if not isinstance(instance, AlarmInstanceData):
            raise TypeError('{} is not of type AlarmInstanceData'.format(type(instance)))

        data = {
            INSTANCE_ID_KEY: instance.instance_id,
            CREATED_KEY: self._string_to_time(instance.created),
            MODIFIED_KEY: self._string_to_time(instance.modified),
            ATTRIBUTES_KEY: dict()
        }
        for attribute in instance.attributes:
            data[ATTRIBUTES_KEY][attribute.name] = self._string_to_attribute(attribute.value)
        return data

    def _class_to_dict(self, val):
        if not isinstance(val, AlarmClassData):
            raise TypeError('{} is not of type AlarmClassData'.format(type(val)))

        data = {
            CLASS_ID_KEY: val.class_id,
        }
        for instance in val.instances:
            data[instance.instance_id] = self._instance_to_dict(instance)
        return data

    def _device_to_dict(self, val):
        if not isinstance(val, AlarmDeviceData):
            raise TypeError('{} is not of type AlarmDeviceData'.format(type(val)))

        data = {
            DEVICE_ID_KEY: val.device_id,
            CREATED_KEY: self._string_to_time(val.created),
            VERSION_KEY: val.version,
            ME_KEY: dict(),
            MSG_TYPE_KEY: set()
        }
        for class_data in val.classes:
            data[class_data.class_id] = self._class_to_dict(class_data)
        for managed_entity in val.managed_entities:
            data[ME_KEY][managed_entity.class_id] = managed_entity.name

        for msg_type in val.message_types:
            data[MSG_TYPE_KEY].add(msg_type.message_type)

        return data



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import, division
import os
import rlcompleter
from pprint import pprint

import structlog
from twisted.conch import manhole_ssh
from twisted.conch.manhole import ColoredManhole
from twisted.conch.ssh import keys
from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
from twisted.cred.portal import Portal
from twisted.internet import reactor
from six.moves import range

log = structlog.get_logger()


MANHOLE_SERVER_RSA_PRIVATE = './manhole_rsa_key'
MANHOLE_SERVER_RSA_PUBLIC = './manhole_rsa_key.pub'


def get_rsa_keys():
    if not (os.path.exists(MANHOLE_SERVER_RSA_PUBLIC) and \
                    os.path.exists(MANHOLE_SERVER_RSA_PRIVATE)):
        # generate a RSA keypair
        log.info('generate-rsa-keypair')
        from Crypto.PublicKey import RSA
        rsa_key = RSA.generate(1024)
        public_key_str = rsa_key.publickey().exportKey(format='OpenSSH')
        private_key_str = rsa_key.exportKey()

        # save keys for next time
        open(MANHOLE_SERVER_RSA_PUBLIC, 'w+b').write(public_key_str)
        open(MANHOLE_SERVER_RSA_PRIVATE, 'w+b').write(private_key_str)
        log.debug('saved-rsa-keypair', public=MANHOLE_SERVER_RSA_PUBLIC,
                  private=MANHOLE_SERVER_RSA_PRIVATE)
    else:
        public_key_str = open(MANHOLE_SERVER_RSA_PUBLIC).read()
        private_key_str = open(MANHOLE_SERVER_RSA_PRIVATE).read()
    return public_key_str, private_key_str


class ManholeWithCompleter(ColoredManhole):

    def __init__(self, namespace):
        namespace['manhole'] = self
        super(ManholeWithCompleter, self).__init__(namespace)
        self.last_tab = None
        self.completer = rlcompleter.Completer(self.namespace)

    def handle_TAB(self):
        if self.last_tab != self.lineBuffer:
            self.last_tab = self.lineBuffer
            return

        buffer = ''.join(self.lineBuffer)
        completions = []
        maxlen = 3
        for c in range(1000):
            candidate = self.completer.complete(buffer, c)
            if not candidate:
                break

            if len(candidate) > maxlen:
                maxlen = len(candidate)

            completions.append(candidate)

        if len(completions) == 1:
            rest = completions[0][len(buffer):]
            self.terminal.write(rest)
            self.lineBufferIndex += len(rest)
            self.lineBuffer.extend(rest)

        elif len(completions):
            maxlen += 3
            numcols = self.width / maxlen
            self.terminal.nextLine()
            for idx, candidate in enumerate(completions):
                self.terminal.write('%%-%ss' % maxlen % candidate)
                if not ((idx + 1) % numcols):
                    self.terminal.nextLine()
            self.terminal.nextLine()
            self.drawInputLine()


class Manhole(object):

    def __init__(self, port, pws, **kw):
        kw.update(globals())
        kw['pp'] = pprint

        realm = manhole_ssh.TerminalRealm()
        manhole = ManholeWithCompleter(kw)

        def windowChanged(_, win_size):
            manhole.terminalSize(*reversed(win_size[:2]))

        realm.sessionFactory.windowChanged = windowChanged
        realm.chainedProtocolFactory.protocolFactory = lambda _: manhole
        portal = Portal(realm)
        portal.registerChecker(InMemoryUsernamePasswordDatabaseDontUse(**pws))
        factory = manhole_ssh.ConchFactory(portal)
        public_key_str, private_key_str = get_rsa_keys()
        factory.publicKeys = {
            'ssh-rsa': keys.Key.fromString(public_key_str)
        }
        factory.privateKeys = {
            'ssh-rsa': keys.Key.fromString(private_key_str)
        }
        reactor.listenTCP(port, factory, interface='localhost')


if __name__ == '__main__':
    Manhole(12222, dict(admin='admin'))
    reactor.run()



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import absolute_import
import arrow
from unittest import TestCase, main
from pyvoltha.adapters.kafka.core_proxy import CoreProxy
from pyvoltha.adapters.extensions.events.adapter_events import AdapterEvents
from pyvoltha.adapters.extensions.events.device_events.onu.onu_activation_fail_event import OnuActivationFailEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_active_event import OnuActiveEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_discovery_event import OnuDiscoveryEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_dying_gasp_event import OnuDyingGaspEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_equipment_event import OnuEquipmentEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_high_rx_optical_power_event import OnuHighRxOpticalEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_high_tx_optical_power_event import OnuHighTxOpticalEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_laser_bias_current_event import OnuLaserBiasEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_laser_eol_event import OnuLaserEolEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_lob_event import OnuLobEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_lopc_mic_error_event import OnuLopcMicErrorEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_lopc_miss_event import OnuLopcMissEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_los_event import OnuLosEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_low_rx_optical_power_event import OnuLowRxOpticalEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_low_tx_optical_power_event import OnuLowTxOpticalEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_selftest_failure_event import OnuSelfTestFailureEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_signal_degrade_event import OnuSignalDegradeEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_signal_fail_event import OnuSignalFailEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_startup_event import OnuStartupEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_temp_red_event import OnuTempRedEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_temp_yellow_event import OnuTempYellowEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_voltage_red_event import OnuVoltageRedEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_voltage_yellow_event import OnuVoltageYellowEvent
from pyvoltha.adapters.extensions.events.device_events.onu.onu_window_drift_event import OnuWindowDriftEvent

DEFAULT_ONU_DEVICE_ID = 'default_onu_mock'
DEFAULT_PON_ID = 0
DEFAULT_ONU_ID = 0
DEFAULT_ONU_SN = 'TEST00000001'
DEFAULT_OLT_SN = 'ABCDXXXXYYYY'
DEFAULT_ONU_REG = 'ABCD1234'

core_proxy = CoreProxy(
               kafka_proxy=None,
               default_core_topic='rwcore',
               default_event_topic='voltha.events',
               my_listening_topic='openonu')

event_mgr = AdapterEvents(core_proxy, DEFAULT_ONU_DEVICE_ID, DEFAULT_ONU_DEVICE_ID, DEFAULT_ONU_SN)


class TestOnuActivationFailEvent(TestCase):

    def setUp(self):
        self.event = OnuActivationFailEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                            arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuActiveEvent(TestCase):

    def setUp(self):
        self.event = OnuActiveEvent(event_mgr, DEFAULT_ONU_DEVICE_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                    DEFAULT_ONU_REG, DEFAULT_OLT_SN, arrow.utcnow().timestamp, onu_id=DEFAULT_ONU_ID )

    def test_get_context_data(self):

        expected_dict = {
            'pon-id': DEFAULT_PON_ID,
            'onu-id': DEFAULT_ONU_ID,
            'serial-number': DEFAULT_ONU_SN,
            'olt_serial_number': DEFAULT_OLT_SN,
            'device_id': DEFAULT_ONU_DEVICE_ID,
            'registration_id': DEFAULT_ONU_REG
        }

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuDiscoveryEvent(TestCase):

        def setUp(self):
            self.event = OnuDiscoveryEvent(event_mgr, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                               arrow.utcnow().timestamp)
        def test_get_context_data(self):
            expected_dict = {'pon-id': DEFAULT_PON_ID,
                             'serial-number': DEFAULT_ONU_SN,
                             'device-type': 'onu'}

            self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuDyingGaspEvent(TestCase):

    def setUp(self):
        self.event = OnuDyingGaspEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuEquipmentEvent(TestCase):

    def setUp(self):
        self.event = OnuEquipmentEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuHighRxOpticalEvent(TestCase):

    def setUp(self):
        self.event = OnuHighRxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuHighTxOpticalEvent(TestCase):

    def setUp(self):
        self.event = OnuHighTxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLaserBiasEvent(TestCase):

    def setUp(self):
        self.event = OnuLaserBiasEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLaserEolEvent(TestCase):

    def setUp(self):
        self.event = OnuLaserEolEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLobEvent(TestCase):

    def setUp(self):
        self.event = OnuLobEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLopcMicErrorEvent(TestCase):

    def setUp(self):
        self.event = OnuLopcMicErrorEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLopcMissEvent(TestCase):

    def setUp(self):
        self.event = OnuLopcMissEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLosEvent(TestCase):

    def setUp(self):
        self.event = OnuLosEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLowRxOpticalEvent(TestCase):

    def setUp(self):
        self.event = OnuLowRxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuLowTxOpticalEvent(TestCase):

    def setUp(self):
        self.event = OnuLowTxOpticalEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuSelfTestFailureEvent(TestCase):

    def setUp(self):
        self.event = OnuSelfTestFailureEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuSignalDegradeEvent(TestCase):

    def setUp(self):
        self.event = OnuSignalDegradeEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, 20, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN,
                         'inverse-bit-error-rate': 20}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuSignalFailEvent(TestCase):

    def setUp(self):
        self.event = OnuSignalFailEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, 20, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN,
                         'inverse-bit-error-rate': 20}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuStartupEvent(TestCase):

    def setUp(self):
        self.event = OnuStartupEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuTempRedEvent(TestCase):

    def setUp(self):
        self.event = OnuTempRedEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuTempYellowEvent(TestCase):

    def setUp(self):
        self.event = OnuTempYellowEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuVoltageRedEvent(TestCase):

    def setUp(self):
        self.event = OnuVoltageRedEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                       arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuVoltageYellowEvent(TestCase):

    def setUp(self):
        self.event = OnuVoltageYellowEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN}

        self.assertEqual(self.event.get_context_data(), expected_dict)

class TestOnuWindowDriftEvent(TestCase):

    def setUp(self):
        self.event = OnuWindowDriftEvent(event_mgr, DEFAULT_ONU_ID, DEFAULT_PON_ID, 10, 20, DEFAULT_ONU_SN,
                                           arrow.utcnow().timestamp)

    def test_get_context_data(self):
        expected_dict = {'onu-id': DEFAULT_ONU_ID,
                         'onu-intf-id': DEFAULT_PON_ID,
                         'onu-serial-number': DEFAULT_ONU_SN,
                         'drift': 10,
                         'new-eqd': 20}

        self.assertEqual(self.event.get_context_data(), expected_dict)


if __name__ == '__main__':
    main()



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Agent to play gateway between adapters.
"""

from __future__ import absolute_import
import structlog
from uuid import uuid4
from twisted.internet.defer import inlineCallbacks, returnValue
from .container_proxy import ContainerProxy
from voltha_protos.inter_container_pb2 import InterAdapterHeader, \
    InterAdapterMessage
import time
import codecs
import six

log = structlog.get_logger()


class AdapterProxy(ContainerProxy):

    def __init__(self, kafka_proxy, core_topic, my_listening_topic):
        super(AdapterProxy, self).__init__(kafka_proxy,
                                           core_topic,
                                           my_listening_topic)

    def _to_string(self, unicode_str):
        if unicode_str is not None:
            if isinstance(unicode_str, six.string_types):
                return unicode_str
            else:
                return codecs.encode(unicode_str, 'ascii')
        else:
            return None

    @ContainerProxy.wrap_request(None)
    @inlineCallbacks
    def send_inter_adapter_message(self,
                                   msg,
                                   type,
                                   from_adapter,
                                   to_adapter,
                                   to_device_id=None,
                                   proxy_device_id=None,
                                   message_id=None):
        """
        Sends a message directly to an adapter. This is typically used to send
        proxied messages from one adapter to another.  An initial ACK response
        is sent back to the invoking adapter.  If there is subsequent response
        to be sent back (async) then the adapter receiving this request will
        use this same API to send back the async response.
        :param msg : GRPC message to send
        :param type : InterAdapterMessageType of the message to send
        :param from_adapter: Name of the adapter making the request.
        :param to_adapter: Name of the remote adapter.
        :param to_device_id: The ID of the device for to the message is
        intended. if it's None then the message is not intended to a specific
        device.  Its interpretation is adapter specific.
        :param proxy_device_id: The ID of the device which will proxy that
        message. If it's None then there is no specific device to proxy the
        message.  Its interpretation is adapter specific.
        :param message_id: A unique number for this transaction that the
        adapter may use to correlate a request and an async response.
        """

        try:
            # validate params
            assert msg
            assert from_adapter
            assert to_adapter

            # Build the inter adapter message
            h = InterAdapterHeader()
            h.type = type
            h.from_topic = self._to_string(from_adapter)
            h.to_topic = self._to_string(to_adapter)
            h.to_device_id = self._to_string(to_device_id)
            h.proxy_device_id = self._to_string(proxy_device_id)

            if message_id:
                h.id = self._to_string(message_id)
            else:
                h.id = uuid4().hex

            h.timestamp = int(round(time.time() * 1000))
            iaMsg = InterAdapterMessage()
            iaMsg.header.CopyFrom(h)
            iaMsg.body.Pack(msg)

            log.debug("sending-inter-adapter-message", type=iaMsg.header.type, from_topic=iaMsg.header.from_topic,
                  to_topic=iaMsg.header.to_topic, to_device_id=iaMsg.header.to_device_id)
            res = yield self.invoke(rpc="process_inter_adapter_message",
                                    to_topic=iaMsg.header.to_topic,
                                    msg=iaMsg)
            returnValue(res)
        except Exception as e:
            log.exception("error-sending-request", e=e)



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from scapy.fields import ByteField, PacketField, IntField
from scapy.fields import ShortField, ConditionalField
from scapy.packet import Packet

from pyvoltha.adapters.extensions.omci.omci_fields import FixedLenField
from pyvoltha.adapters.extensions.omci.omci_messages import OmciCreate, OmciDelete, \
    OmciDeleteResponse, OmciSet, OmciSetResponse, OmciGet, OmciGetResponse, \
    OmciGetAllAlarms, OmciGetAllAlarmsResponse, OmciGetAllAlarmsNext, \
    OmciMibResetResponse, OmciMibReset, OmciMibUploadNextResponse, \
    OmciMibUploadNext, OmciMibUploadResponse, OmciMibUpload, \
    OmciGetAllAlarmsNextResponse, OmciAttributeValueChange, \
    OmciTestResult, OmciAlarmNotification, \
    OmciReboot, OmciRebootResponse, OmciGetNext, OmciGetNextResponse, \
    OmciSynchronizeTime, OmciSynchronizeTimeResponse, OmciGetCurrentData, \
    OmciGetCurrentDataResponse, OmciStartSoftwareDownload, OmciStartSoftwareDownloadResponse, \
    OmciDownloadSection, OmciDownloadSectionLast, OmciDownloadSectionResponse, \
    OmciEndSoftwareDownload, OmciEndSoftwareDownloadResponse, \
    OmciActivateImage, OmciActivateImageResponse, \
    OmciCommitImage, OmciCommitImageResponse, OmciCreateResponse, OmciTestResponse, OmciTest


class OmciFrame(Packet):
    name = "OmciFrame"
    fields_desc = [
        ShortField("transaction_id", 0),
        ByteField("message_type", None),
        ByteField("omci", 0x0a),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciCreate), align=36),
            lambda pkt: pkt.message_type == OmciCreate.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciCreateResponse), align=36),
            lambda pkt: pkt.message_type == OmciCreateResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciDelete), align=36),
            lambda pkt: pkt.message_type == OmciDelete.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciDeleteResponse), align=36),
            lambda pkt: pkt.message_type == OmciDeleteResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciSet), align=36),
            lambda pkt: pkt.message_type == OmciSet.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciSetResponse), align=36),
            lambda pkt: pkt.message_type == OmciSetResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGet), align=36),
            lambda pkt: pkt.message_type == OmciGet.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetResponse), align=36),
            lambda pkt: pkt.message_type == OmciGetResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetAllAlarms), align=36),
            lambda pkt: pkt.message_type == OmciGetAllAlarms.message_id),
        ConditionalField(FixedLenField(
            PacketField(
                "omci_message", None, OmciGetAllAlarmsResponse), align=36),
                lambda pkt:
                pkt.message_type == OmciGetAllAlarmsResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetAllAlarmsNext), align=36),
            lambda pkt: pkt.message_type == OmciGetAllAlarmsNext.message_id),
        ConditionalField(FixedLenField(
            PacketField(
                "omci_message", None, OmciGetAllAlarmsNextResponse), align=36),
                lambda pkt:
                pkt.message_type == OmciGetAllAlarmsNextResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciMibUpload), align=36),
            lambda pkt: pkt.message_type == OmciMibUpload.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciMibUploadResponse), align=36),
            lambda pkt: pkt.message_type == OmciMibUploadResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciMibUploadNext), align=36),
            lambda pkt:
                pkt.message_type == OmciMibUploadNext.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciMibUploadNextResponse), align=36),
            lambda pkt: pkt.message_type == OmciMibUploadNextResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciMibReset), align=36),
            lambda pkt: pkt.message_type == OmciMibReset.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciMibResetResponse), align=36),
            lambda pkt: pkt.message_type == OmciMibResetResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciAlarmNotification), align=36),
            lambda pkt: pkt.message_type == OmciAlarmNotification.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciAttributeValueChange), align=36),
            lambda pkt: pkt.message_type == OmciAttributeValueChange.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciReboot), align=36),
            lambda pkt: pkt.message_type == OmciReboot.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciRebootResponse), align=36),
            lambda pkt: pkt.message_type == OmciRebootResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetNext), align=36),
            lambda pkt: pkt.message_type == OmciGetNext.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetNextResponse), align=36),
            lambda pkt: pkt.message_type == OmciGetNextResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciSynchronizeTime), align=36),
            lambda pkt: pkt.message_type == OmciSynchronizeTime.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciSynchronizeTimeResponse), align=36),
            lambda pkt: pkt.message_type == OmciSynchronizeTimeResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetCurrentData), align=36),
            lambda pkt: pkt.message_type == OmciGetCurrentData.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciGetCurrentDataResponse), align=36),
            lambda pkt: pkt.message_type == OmciGetCurrentDataResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciStartSoftwareDownload), align=36),
            lambda pkt: pkt.message_type == OmciStartSoftwareDownload.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciStartSoftwareDownloadResponse), align=36),
            lambda pkt: pkt.message_type == OmciStartSoftwareDownloadResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciDownloadSection), align=36),
            lambda pkt: pkt.message_type == OmciDownloadSection.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciDownloadSectionLast), align=36),
            lambda pkt: pkt.message_type == OmciDownloadSectionLast.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciDownloadSectionResponse), align=36),
            lambda pkt: pkt.message_type == OmciDownloadSectionResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciEndSoftwareDownload), align=36),
            lambda pkt: pkt.message_type == OmciEndSoftwareDownload.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciEndSoftwareDownloadResponse), align=36),
            lambda pkt: pkt.message_type == OmciEndSoftwareDownloadResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciActivateImage), align=36),
            lambda pkt: pkt.message_type == OmciActivateImage.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciActivateImageResponse), align=36),
            lambda pkt: pkt.message_type == OmciActivateImageResponse.message_id),

        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciCommitImage), align=36),
            lambda pkt: pkt.message_type == OmciCommitImage.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciCommitImageResponse), align=36),
            lambda pkt: pkt.message_type == OmciCommitImageResponse.message_id),
        # Create Frame for Omci Test.
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciTest), align=36),
            lambda pkt: pkt.message_type == OmciTest.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciTestResponse), align=36),
            lambda pkt: pkt.message_type == OmciTestResponse.message_id),
        ConditionalField(FixedLenField(
            PacketField("omci_message", None, OmciTestResult), align=36),
            lambda pkt: pkt.message_type == OmciTestResult.message_id),

        # TODO add entries for remaining OMCI message types

        IntField("omci_trailer", 0x00000028)
    ]

    # We needed to patch the do_dissect(...) method of Packet, because
    # it wiped out already dissected conditional fields with None if they
    # referred to the same field name. We marked the only new line of code
    # with "Extra condition added".
    def do_dissect(self, s):
        raw = s
        self.raw_packet_cache_fields = {}
        for f in self.fields_desc:
            if not s:
                break
            s, fval = f.getfield(self, s)
            # We need to track fields with mutable values to discard
            # .raw_packet_cache when needed.
            if f.islist or f.holds_packets:
                self.raw_packet_cache_fields[f.name] = f.do_copy(fval)
            # Extra condition added
            if fval is not None or f.name not in self.fields:
                self.fields[f.name] = fval
        assert(raw.endswith(s))
        self.raw_packet_cache = raw[:-len(s)] if s else raw
        self.explicit = 1
        return s



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
import structlog
from scapy.fields import ByteField, ThreeBytesField, StrFixedLenField, ConditionalField, IntField, Field
from scapy.fields import ShortField, BitField
from scapy.packet import Packet

from pyvoltha.adapters.extensions.omci.omci_defs import AttributeAccess, OmciSectionDataSize
from pyvoltha.adapters.extensions.omci.omci_fields import OmciTableField, OmciVariableLenZeroPadField
import pyvoltha.adapters.extensions.omci.omci_entities as omci_entities


log = structlog.get_logger()


class OmciData(Field):

    __slots__ = Field.__slots__ + ['_entity_class']

    def __init__(self, name, entity_class="entity_class"):
        Field.__init__(self, name=name, default=None, fmt='s')
        self._entity_class = entity_class

    def addfield(self, pkt, s, val):
        class_id = getattr(pkt, self._entity_class)
        entity_class = omci_entities.entity_id_to_class_map.get(class_id)
        for attribute in entity_class.attributes:
            if AttributeAccess.SetByCreate not in attribute.access:
                continue
            if attribute.field.name == 'managed_entity_id':
                continue
            fld = attribute.field
            s = fld.addfield(pkt, s, val.get(fld.name, fld.default))
        return s

    def getfield(self, pkt, s):
        """Extract an internal value from a string"""
        class_id = getattr(pkt, self._entity_class)
        entity_class = omci_entities.entity_id_to_class_map.get(class_id)
        data = {}
        for attribute in entity_class.attributes:
            if AttributeAccess.SetByCreate not in attribute.access:
                continue
            if attribute.field.name == 'managed_entity_id':
                continue
            fld = attribute.field
            s, value = fld.getfield(pkt, s)
            data[fld.name] = value
        return s, data


class OmciMaskedData(Field):

    __slots__ = Field.__slots__ + ['_entity_class', '_attributes_mask']

    def __init__(self, name, entity_class="entity_class",
                 attributes_mask="attributes_mask"):
        Field.__init__(self, name=name, default=None, fmt='s')
        self._entity_class = entity_class
        self._attributes_mask = attributes_mask

    def addfield(self, pkt, s, val):
        class_id = getattr(pkt, self._entity_class)
        attribute_mask = getattr(pkt, self._attributes_mask)
        entity_class = omci_entities.entity_id_to_class_map.get(class_id)
        indices = entity_class.attribute_indices_from_mask(attribute_mask)
        for index in indices:
            fld = entity_class.attributes[index].field
            s = fld.addfield(pkt, s, val[fld.name])
        return s

    def getfield(self, pkt, s):
        """Extract an internal value from a string"""
        class_id = getattr(pkt, self._entity_class)
        attribute_mask = getattr(pkt, self._attributes_mask)
        entity_class = omci_entities.entity_id_to_class_map[class_id]
        indices = entity_class.attribute_indices_from_mask(attribute_mask)
        data = {}
        table_attribute_mask = 0
        for index in indices:
            try:
                fld = entity_class.attributes[index].field
            except IndexError as e:
                log.error("attribute-decode-failure", attribute_index=index,
                          entity_class=entity_class, e=e)
                continue
            try:
                s, value = fld.getfield(pkt, s)
            except Exception as _e:
                raise
            if isinstance(pkt, OmciGetResponse) and isinstance(fld, OmciTableField):
                data[fld.name + '_size'] = value
                table_attribute_mask = table_attribute_mask | (1 << (16 - index))
            else:
                data[fld.name] = value
        if table_attribute_mask:
            data['table_attribute_mask'] = table_attribute_mask
        return s, data


class OmciMessage(Packet):
    name = "OmciMessage"
    message_id = None  # OMCI message_type value, filled by derived classes
    fields_desc = []


class OmciCreate(OmciMessage):
    name = "OmciCreate"
    message_id = 0x44
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        OmciData("data")
    ]


class OmciCreateResponse(OmciMessage):
    name = "OmciCreateResponse"
    message_id = 0x24
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", None),
        ByteField("success_code", 0),
        ShortField("parameter_error_attributes_mask", None),
    ]


class OmciDelete(OmciMessage):
    name = "OmciDelete"
    message_id = 0x46
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", None),
    ]


class OmciDeleteResponse(OmciMessage):
    name = "OmciDeleteResponse"
    message_id = 0x26
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", None),
        ByteField("success_code", 0),
    ]


class OmciSet(OmciMessage):
    name = "OmciSet"
    message_id = 0x48
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ShortField("attributes_mask", None),
        OmciMaskedData("data")
    ]


class OmciSetResponse(OmciMessage):
    name = "OmciSetResponse"
    message_id = 0x28
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", None),
        ByteField("success_code", 0),
        ShortField("unsupported_attributes_mask", None),
        ShortField("failed_attributes_mask", None),
    ]


class OmciGet(OmciMessage):
    name = "OmciGet"
    message_id = 0x49
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ShortField("attributes_mask", None)
    ]


class OmciGetResponse(OmciMessage):
    name = "OmciGetResponse"
    message_id = 0x29
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("success_code", 0),
        ShortField("attributes_mask", None),
        ConditionalField(OmciMaskedData("data"),
                         lambda pkt: pkt.success_code in (0, 9)),
        ConditionalField(OmciVariableLenZeroPadField("zero_padding", 36),
                         lambda pkt: pkt.success_code == 9),

        # These fields are only valid if attribute error (status == 9)
        ConditionalField(ShortField("unsupported_attributes_mask", 0),
                         lambda pkt: pkt.success_code == 9),
        ConditionalField(ShortField("failed_attributes_mask", 0),
                         lambda pkt: pkt.success_code == 9)
    ]


class OmciGetAllAlarms(OmciMessage):
    name = "OmciGetAllAlarms"
    message_id = 0x4b
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),  # Always 0 (ONT instance)
        ByteField("alarm_retrieval_mode", 0)  # 0 or 1
    ]


class OmciGetAllAlarmsResponse(OmciMessage):
    name = "OmciGetAllAlarmsResponse"
    message_id = 0x2b
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
        ShortField("number_of_commands", None)
    ]


class OmciGetAllAlarmsNext(OmciMessage):
    name = "OmciGetAllAlarmsNext"
    message_id = 0x4c
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
        ShortField("command_sequence_number", None)
    ]


class OmciGetAllAlarmsNextResponse(OmciMessage):
    name = "OmciGetAllAlarmsNextResponse"
    message_id = 0x2c
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
        ShortField("alarmed_entity_class", None),
        ShortField("alarmed_entity_id", 0),
        BitField("alarm_bit_map", None, 224)
    ]


class OmciMibUpload(OmciMessage):
    name = "OmciMibUpload"
    message_id = 0x4d
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
    ]


class OmciMibUploadResponse(OmciMessage):
    name = "OmciMibUploadResponse"
    message_id = 0x2d
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
        ShortField("number_of_commands", None)
    ]


class OmciMibUploadNext(OmciMessage):
    name = "OmciMibUploadNext"
    message_id = 0x4e
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
        ShortField("command_sequence_number", None)
    ]


class OmciMibUploadNextResponse(OmciMessage):
    name = "OmciMibUploadNextResponse"
    message_id = 0x2e
    fields_desc = [
        ShortField("entity_class", 2),  # Always 2 (ONT data)
        ShortField("entity_id", 0),
        ShortField("object_entity_class", None),
        ShortField("object_entity_id", 0),
        ShortField("object_attributes_mask", None),
        OmciMaskedData("object_data", entity_class='object_entity_class',
                       attributes_mask='object_attributes_mask')
    ]


class OmciMibReset(OmciMessage):
    name = "OmciMibReset"
    message_id = 0x4f
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0)
    ]


class OmciMibResetResponse(OmciMessage):
    name = "OmciMibResetResponse"
    message_id = 0x2f
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("success_code", 0)
    ]


class OmciAlarmNotification(OmciMessage):
    name = "AlarmNotification"
    message_id = 0x10
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        BitField("alarm_bit_map", 0, 224),
        ThreeBytesField("zero_padding", 0),
        ByteField("alarm_sequence_number", None)
    ]


class OmciAttributeValueChange(OmciMessage):
    name = "AttributeValueChange"
    message_id = 0x11
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ShortField("attributes_mask", None),
        OmciMaskedData("data")
    ]

class OmciReboot(OmciMessage):
    name = "OmciOnuReboot"
    message_id = 0x59
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("reboot_code", 0)
    ]


class OmciRebootResponse(OmciMessage):
    name = "OmciOnuRebootResponse"
    message_id = 0x39
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("success_code", 0)
    ]


class OmciGetNext(OmciMessage):
    name = "OmciGetNext"
    message_id = 0x5A
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ShortField("attributes_mask", None),
        ShortField("command_sequence_number", None)
    ]


class OmciGetNextResponse(OmciMessage):
    name = "OmciGetNextResponse"
    message_id = 0x3A
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("success_code", 0),
        ShortField("attributes_mask", None),
        ConditionalField(OmciMaskedData("data"),
                         lambda pkt: pkt.success_code == 0)
    ]


class OmciSynchronizeTime(OmciMessage):
    name = "OmciSynchronizeTime"
    message_id = 0x58
    fields_desc = [
        ShortField("entity_class", 256),  # OntG
        ShortField("entity_id", 0),
        ShortField("year", 0),       # eg) 2018
        ByteField("month", 0),       # 1..12
        ByteField("day", 0),         # 1..31
        ByteField("hour", 0),        # 0..23
        ByteField("minute", 0),      # 0..59
        ByteField("second", 0)       # 0..59
    ]


class OmciSynchronizeTimeResponse(OmciMessage):
    name = "OmciSynchronizeTimeResponse"
    message_id = 0x38
    fields_desc = [
        ShortField("entity_class", 256),  # OntG
        ShortField("entity_id", 0),
        ByteField("success_code", 0),
        ConditionalField(ShortField("success_info", None),
                         lambda pkt: pkt.success_code == 0)
    ]


class OmciGetCurrentData(OmciMessage):
    name = "OmciGetCurrentData"
    message_id = 0x5C
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ShortField("attributes_mask", None),
    ]


class OmciGetCurrentDataResponse(OmciMessage):
    name = "OmciGetCurrentDataResponse"
    message_id = 0x3C
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("success_code", 0),
        ShortField("attributes_mask", None),
        ShortField("unsupported_attributes_mask", None),
        ShortField("failed_attributes_mask", None),
        ConditionalField(
            OmciMaskedData("data"), lambda pkt: pkt.success_code == 0)
    ]


class OmciStartSoftwareDownload(OmciMessage):
    name = "OmciStartSoftwareDownload"
    message_id = 0x53
    fields_desc = [
        ShortField("entity_class", 7),  # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("window_size", 0),
        IntField("image_size", 0),
        ByteField("image_number", 1),   # Always only 1 in parallel
        ShortField("instance_id", None) # should be same as "entity_id"        
    ]


class OmciStartSoftwareDownloadResponse(OmciMessage):
    name = "OmciStartSoftwareDownloadResponse"
    message_id = 0x33
    fields_desc = [
        ShortField("entity_class", 7),  # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("result", 0),
        ByteField("window_size", 0),
        ByteField("image_number", 1),   # Always only 1 in parallel
        ShortField("instance_id", None) # should be same as "entity_id"        
    ]


class OmciEndSoftwareDownload(OmciMessage):
    name = "OmciEndSoftwareDownload"
    message_id = 0x55
    fields_desc = [
        ShortField("entity_class", 7),  # Always 7 (Software image)
        ShortField("entity_id", None),
        IntField("crc32", 0),
        IntField("image_size", 0),
        ByteField("image_number", 1),   # Always only 1 in parallel
        ShortField("instance_id", None),# should be same as "entity_id"
    ]


class OmciEndSoftwareDownloadResponse(OmciMessage):
    name = "OmciEndSoftwareDownload"
    message_id = 0x35
    fields_desc = [
        ShortField("entity_class", 7),  # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("result", 0),
        ByteField("image_number", 1),   # Always only 1 in parallel
        ShortField("instance_id", None),# should be same as "entity_id"
        ByteField("result0", 0)         # same as result 
    ]


class OmciDownloadSection(OmciMessage):
    name = "OmciDownloadSection"
    message_id = 0x14
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("section_number", 0),  # Always only 1 in parallel
        StrFixedLenField("data", 0, length=OmciSectionDataSize) # section data
    ]


class OmciDownloadSectionLast(OmciMessage):
    name = "OmciDownloadSection"
    message_id = 0x54
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("section_number", 0),  # Always only 1 in parallel
        StrFixedLenField("data", 0, length=OmciSectionDataSize) # section data
    ]


class OmciDownloadSectionResponse(OmciMessage):
    name = "OmciDownloadSectionResponse"
    message_id = 0x34
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("result", 0),
        ByteField("section_number", 0),  # Always only 1 in parallel
    ]


class OmciActivateImage(OmciMessage):
    name = "OmciActivateImage"
    message_id = 0x56
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("activate_flag", 0)    # Activate image unconditionally
    ]


class OmciActivateImageResponse(OmciMessage):
    name = "OmciActivateImageResponse"
    message_id = 0x36
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("result", 0)           # Activate image unconditionally
    ]


class OmciCommitImage(OmciMessage):
    name = "OmciCommitImage"
    message_id = 0x57
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
    ]


class OmciCommitImageResponse(OmciMessage):
    name = "OmciCommitImageResponse"
    message_id = 0x37
    fields_desc = [
        ShortField("entity_class", 7),   # Always 7 (Software image)
        ShortField("entity_id", None),
        ByteField("result", 0)           # Activate image unconditionally
    ]

class OmciTest(OmciMessage):
    name = "OmciTest"
    message_id = 0x52
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ShortField('self_test', 0x07)
    ]


class OmciTestResponse(OmciMessage):
    name = "OmciTesResponse"
    message_id = 0x32
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("success_code", None)
    ]

class OmciTestResult(OmciMessage):
    name = "TestResult"
    message_id = 0x1B
    fields_desc = [
        ShortField("entity_class", None),
        ShortField("entity_id", 0),
        ByteField("power_feed_type_id", 1),
        ShortField("power_feed_voltage", None),
        ByteField("received_optical_power_type_id", 3),
        ShortField('received_optical_power', None),
        ByteField("mean_optical_launch_power_type_id", 5),
        ShortField('mean_optical_launch_power', None),
        ByteField("laser_bias_current_type_id", 9),
        ShortField('laser_bias_current', None),
        ByteField("temperature_type_id", 12),
        ShortField('temperature', None)
    ]




# Copyright 2017-present Adtran, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from voltha_protos.events_pb2 import AlarmEventType, AlarmEventSeverity, AlarmEventCategory
from pyvoltha.adapters.extensions.alarms.adapter_alarms import AlarmBase


class OnuVoltageRedAlarm(AlarmBase):
    """
    The ONU Voltage Red Alarm is reported by the ONT-G (ME # 256) to
    indicate some services have been shut down to avoid power collapse.
    The operational state of the affected PPTPs indicates the affected
    services.

    For ONT-G equipment alarms, the intf_id reported is that of the PON/ANI
    physical port number
    """
    def __init__(self, alarm_mgr, onu_id, intf_id, serial_number):
        super(OnuVoltageRedAlarm, self).__init__(alarm_mgr, object_type='onu voltage red',
                                                 alarm='ONU_VOLTAGE_RED',
                                                 alarm_category=AlarmEventCategory.ONU,
                                                 alarm_type=AlarmEventType.COMMUNICATION,
                                                 alarm_severity=AlarmEventSeverity.CRITICAL)
        self._onu_id = onu_id
        self._intf_id = intf_id
        self._serial_number = serial_number

    def get_context_data(self):
        return {'onu-id': self._onu_id,
                'onu-intf-id': self._intf_id,
                'onu-serial-number': self._serial_number}



# Copyright 2017-present Adtran, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import
from voltha_protos.events_pb2 import AlarmEventType, AlarmEventSeverity, AlarmEventCategory
from pyvoltha.adapters.extensions.alarms.adapter_alarms import AlarmBase

"""
    OnuSignalsFailureIndication {
                fixed32 intf_id = 1;
                fixed32 onu_id = 2;
                string status = 3;
                fixed32 inverse_bit_error_rate = 4;
"""

class OnuSignalFailAlarm(AlarmBase):
    def __init__(self, alarm_mgr, onu_id, intf_id, inverse_bit_error_rate, serial_number):
        super(OnuSignalFailAlarm, self).__init__(alarm_mgr, object_type='onu SIGNAL FAIL',
                                          alarm='ONU_SIGNAL_FAIL',
                                          alarm_category=AlarmEventCategory.ONU,
                                          alarm_type=AlarmEventType.COMMUNICATION,
                                          alarm_severity=AlarmEventSeverity.MAJOR)
        self._onu_id = onu_id
        self._intf_id = intf_id
        self._inverse_bit_error_rate = inverse_bit_error_rate
        self._serial_number = serial_number

    def get_context_data(self):
        return {'onu-id': self._onu_id,
                'onu-intf-id': self._intf_id,
                'inverse-bit-error-rate': self._inverse_bit_error_rate,
                 'onu-serial-number': self._serial_number}



#
# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from .task import Task
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, failure
from pyvoltha.adapters.extensions.omci.omci_defs import ReasonCodes, EntityOperations
from pyvoltha.adapters.extensions.omci.omci_frame import OmciFrame
from pyvoltha.adapters.extensions.omci.omci_messages import OmciDelete

RC = ReasonCodes
OP = EntityOperations


class DeletePMException(Exception):
    pass


class OmciDeletePMRequest(Task):
    """
    OpenOMCI routine to delete the requested PM Interval MEs
    """
    task_priority = Task.DEFAULT_PRIORITY
    name = "ONU OMCI Delete PM ME Task"

    def __init__(self, omci_agent, device_id, me_set, exclusive=False):
        """
        Class initialization

        :param omci_agent: (OmciAdapterAgent) OMCI Adapter agent
        :param me_set: (set) Tuples of class_id / entity_id to create
        :param exclusive: (bool) True if this Create request Task exclusively own the
                                 OMCI-CC while running. Default: False
        """
        super(OmciDeletePMRequest, self).__init__(OmciDeletePMRequest.name,
                                                  omci_agent,
                                                  device_id,
                                                  priority=OmciDeletePMRequest.task_priority,
                                                  exclusive=exclusive)
        self._device = omci_agent.get_device(device_id)
        self._me_tuples = me_set
        self._local_deferred = None

    def cancel_deferred(self):
        super(OmciDeletePMRequest, self).cancel_deferred()

        d, self._local_deferred = self._local_deferred, None
        try:
            if d is not None and not d.called:
                d.cancel()
        except:
            pass

    def start(self):
        """ Start task """
        super(OmciDeletePMRequest, self).start()
        self._local_deferred = reactor.callLater(0, self.perform_delete)

    @inlineCallbacks
    def perform_delete(self):
        """ Perform the delete requests """
        self.log.debug('perform-delete')

        try:
            for me in self._me_tuples:
                class_id = me[0]
                entity_id = me[1]

                frame = OmciFrame(
                    transaction_id=None,
                    message_type=OmciDelete.message_id,
                    omci_message=OmciDelete(
                        entity_class=class_id,
                        entity_id=entity_id
                    )
                )
                self.strobe_watchdog()
                results = yield self._device.omci_cc.send(frame)

                status = results.fields['omci_message'].fields['success_code']
                self.log.debug('perform-delete-status', status=status)

                # Did it fail, it instance does not exist, not an error
                if status != RC.Success.value and status != RC.UnknownInstance.value:
                    msg = 'ME: {}, entity: {} failed with status {}'.format(class_id,
                                                                            entity_id,
                                                                            status)
                    raise DeletePMException(msg)

                self.log.debug('delete-pm-success', class_id=class_id,
                               entity_id=entity_id)
            self.deferred.callback(self)

        except Exception as e:
            self.log.exception('perform-create', e=e)
            self.deferred.errback(failure.Failure(e))



#!/usr/bin/env python

# Copyright 2018 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import time
from uuid import uuid4

import structlog
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, returnValue, Deferred, \
    DeferredQueue, gatherResults
from zope.interface import implementer

from pyvoltha.common.utils import asleep
from pyvoltha.common.utils.registry import IComponent
from .kafka_proxy import KafkaProxy, get_kafka_proxy
from voltha_protos.inter_container_pb2 import MessageType, Argument, \
    InterContainerRequestBody, InterContainerMessage, Header, \
    InterContainerResponseBody, StrType
import six
import codecs

log = structlog.get_logger()

KAFKA_OFFSET_LATEST = 'latest'
KAFKA_OFFSET_EARLIEST = 'earliest'
ARG_FROM_TOPIC = 'fromTopic'


class KafkaMessagingError(Exception):
    def __init__(self, error):
        self.error = error


@implementer(IComponent)
class IKafkaMessagingProxy(object):
    _kafka_messaging_instance = None

    def __init__(self,
                 kafka_host_port,
                 kv_store,
                 default_topic,
                 group_id_prefix,
                 target_cls):
        """
        Initialize the kafka proxy.  This is a singleton (may change to
        non-singleton if performance is better)
        :param kafka_host_port: Kafka host and port
        :param kv_store: Key-Value store
        :param default_topic: Default topic to subscribe to
        :param target_cls: target class - method of that class is invoked
        when a message is received on the default_topic
        """
        # return an exception if the object already exist
        if IKafkaMessagingProxy._kafka_messaging_instance:
            raise Exception(
                'Singleton-exist', cls=IKafkaMessagingProxy)

        log.debug("Initializing-KafkaProxy")
        self.kafka_host_port = kafka_host_port
        self.kv_store = kv_store
        self.default_topic = default_topic
        self.default_group_id = "_".join((group_id_prefix, default_topic))
        self.target_cls = target_cls
        self.topic_target_cls_map = {}
        self.topic_callback_map = {}
        self.subscribers = {}
        self.kafka_proxy = None
        self.transaction_id_deferred_map = {}
        self.received_msg_queue = DeferredQueue()
        self.stopped = False

        self.init_time = 0
        self.init_received_time = 0

        self.init_resp_time = 0
        self.init_received_resp_time = 0

        self.num_messages = 0
        self.total_time = 0
        self.num_responses = 0
        self.total_time_responses = 0
        log.debug("KafkaProxy-initialized")

    def start(self):
        try:
            log.debug("KafkaProxy-starting")

            # Get the kafka proxy instance.  If it does not exist then
            # create it
            self.kafka_proxy = get_kafka_proxy()
            if self.kafka_proxy == None:
                KafkaProxy(kafka_endpoint=self.kafka_host_port).start()
                self.kafka_proxy = get_kafka_proxy()

            # Subscribe the default topic and target_cls
            self.topic_target_cls_map[self.default_topic] = self.target_cls

            # Start the queue to handle incoming messages
            reactor.callLater(0, self._received_message_processing_loop)

            # Subscribe using the default topic and default group id.  Whenever
            # a message is received on that topic then teh target_cls will be
            # invoked.
            reactor.callLater(0, self.subscribe,
                              topic=self.default_topic,
                              target_cls=self.target_cls,
                              group_id=self.default_group_id)

            # Setup the singleton instance
            IKafkaMessagingProxy._kafka_messaging_instance = self
            log.debug("KafkaProxy-started")
        except Exception as e:
            log.exception("Failed-to-start-proxy", e=e)

    def stop(self):
        """
        Invoked to stop the kafka proxy
        :return: None on success, Exception on failure
        """
        log.debug("Stopping-messaging-proxy ...")
        try:
            # Stop the kafka proxy.  This will stop all the consumers
            # and producers
            self.stopped = True
            self.kafka_proxy.stop()
            log.debug("Messaging-proxy-stopped.")
        except Exception as e:
            log.exception("Exception-when-stopping-messaging-proxy:", e=e)

    def get_target_cls(self):
        return self.target_cls

    def get_default_topic(self):
        return self.default_topic

    @inlineCallbacks
    def _subscribe_group_consumer(self, group_id, topic, offset, callback=None,
                                  target_cls=None):
        try:
            log.debug("subscribing-to-topic-start", topic=topic)
            yield self.kafka_proxy.subscribe(topic,
                                             self._enqueue_received_group_message,
                                             group_id, offset)

            if target_cls is not None and callback is None:
                # Scenario #1
                if topic not in self.topic_target_cls_map:
                    self.topic_target_cls_map[topic] = target_cls
            elif target_cls is None and callback is not None:
                # Scenario #2
                log.debug("custom-callback", topic=topic,
                          callback_map=self.topic_callback_map)
                if topic not in self.topic_callback_map:
                    self.topic_callback_map[topic] = [callback]
                else:
                    self.topic_callback_map[topic].extend([callback])
            else:
                log.warn("invalid-parameters")

            returnValue(True)
        except Exception as e:
            log.exception("Exception-during-subscription", e=e)
            returnValue(False)

    @inlineCallbacks
    def subscribe(self, topic, callback=None, target_cls=None,
                  max_retry=3, group_id=None, offset=KAFKA_OFFSET_LATEST):
        """
        Scenario 1:  invoked to subscribe to a specific topic with a
        target_cls to invoke when a message is received on that topic.  This
        handles the case of request/response where this library performs the
        heavy lifting. In this case the m_callback must to be None

        Scenario 2:  invoked to subscribe to a specific topic with a
        specific callback to invoke when a message is received on that topic.
        This handles the case where the caller wants to process the message
        received itself. In this case the target_cls must to be None

        :param topic: topic to subscribe to
        :param callback: Callback to invoke when a message is received on
        the topic. Either one of callback or target_cls needs can be none
        :param target_cls:  Target class to use when a message is
        received on the topic. There can only be 1 target_cls per topic.
        Either one of callback or target_cls needs can be none
        :param max_retry:  the number of retries before reporting failure
        to subscribe.  This caters for scenario where the kafka topic is not
        ready.
        :param group_id:  The ID of the group the consumer is subscribing to
        :param offset: The topic offset on the kafka bus from where message consumption will start
        :return: True on success, False on failure
        """
        RETRY_BACKOFF = [0.05, 0.1, 0.2, 0.5, 1, 2, 5]

        def _backoff(msg, retries):
            wait_time = RETRY_BACKOFF[min(retries,
                                          len(RETRY_BACKOFF) - 1)]
            log.info(msg, retry_in=wait_time)
            return asleep.asleep(wait_time)

        log.debug("subscribing", topic=topic, group_id=group_id,
                  callback=callback, target=target_cls)

        retry = 0
        subscribed = False
        if group_id is None:
            group_id = self.default_group_id
        while not subscribed:
            subscribed = yield self._subscribe_group_consumer(group_id, topic,
                                                              callback=callback,
                                                              target_cls=target_cls,
                                                              offset=offset)
            if subscribed:
                returnValue(True)
            elif retry > max_retry:
                returnValue(False)
            else:
                _backoff("subscription-not-complete", retry)
                retry += 1

    def unsubscribe(self, topic, callback=None, target_cls=None):
        """
        Invoked when unsubscribing to a topic
        :param topic: topic to unsubscribe from
        :param callback:  the callback used when subscribing to the topic, if any
        :param target_cls: the targert class used when subscribing to the topic, if any
        :return: None on success or Exception on failure
        """
        log.debug("Unsubscribing-to-topic", topic=topic)

        try:
            self.kafka_proxy.unsubscribe(topic,
                                         self._enqueue_received_group_message)

            if callback is None and target_cls is None:
                log.error("both-call-and-target-cls-cannot-be-none",
                          topic=topic)
                raise KafkaMessagingError(
                    error="both-call-and-target-cls-cannot-be-none")

            if target_cls is not None and topic in self.topic_target_cls_map:
                del self.topic_target_cls_map[topic]

            if callback is not None and topic in self.topic_callback_map:
                index = 0
                for cb in self.topic_callback_map[topic]:
                    if cb == callback:
                        break
                    index += 1
                if index < len(self.topic_callback_map[topic]):
                    self.topic_callback_map[topic].pop(index)

                if len(self.topic_callback_map[topic]) == 0:
                    del self.topic_callback_map[topic]
        except Exception as e:
            log.exception("Exception-when-unsubscribing-to-topic", topic=topic,
                          e=e)
            return e

    @inlineCallbacks
    def _enqueue_received_group_message(self, msg):
        """
        Internal method to continuously queue all received messaged
        irrespective of topic
        :param msg: Received message
        :return: None on success, Exception on failure
        """
        try:
            log.debug("received-msg", msg=msg)
            yield self.received_msg_queue.put(msg)
        except Exception as e:
            log.exception("Failed-enqueueing-received-message", e=e)

    @inlineCallbacks
    def _received_message_processing_loop(self):
        """
        Internal method to continuously process all received messages one
        at a time
        :return: None on success, Exception on failure
        """
        while True:
            try:
                message = yield self.received_msg_queue.get()
                yield self._process_message(message)
                if self.stopped:
                    break
            except Exception as e:
                log.exception("Failed-dequeueing-received-message", e=e)

    def _to_string(self, unicode_str):
        if unicode_str is not None:
            if isinstance(unicode_str, six.string_types):
                return unicode_str
            else:
                return codecs.encode(unicode_str, 'ascii')
        else:
            return None

    def _format_request(self,
                        rpc,
                        to_topic,
                        reply_topic,
                        **kwargs):
        """
        Format a request to send over kafka
        :param rpc: Requested remote API
        :param to_topic: Topic to send the request
        :param reply_topic: Topic to receive the resulting response, if any
        :param kwargs: Dictionary of key-value pairs to pass as arguments to
        the remote rpc API.
        :return: A InterContainerMessage message type on success or None on
        failure
        """
        try:
            transaction_id = uuid4().hex
            request = InterContainerMessage()
            request_body = InterContainerRequestBody()
            request.header.id = transaction_id
            request.header.type = MessageType.Value("REQUEST")
            request.header.from_topic = reply_topic
            request.header.to_topic = to_topic

            response_required = False
            if reply_topic:
                request_body.reply_to_topic = reply_topic
                request_body.response_required = True
                response_required = True

            request.header.timestamp = int(round(time.time() * 1000))
            request_body.rpc = rpc
            for a, b in six.iteritems(kwargs):
                arg = Argument()
                arg.key = a
                try:
                    arg.value.Pack(b)
                    request_body.args.extend([arg])
                except Exception as e:
                    log.exception("Failed-parsing-value", e=e)
            request.body.Pack(request_body)
            return request, transaction_id, response_required
        except Exception as e:
            log.exception("formatting-request-failed",
                          rpc=rpc,
                          to_topic=to_topic,
                          reply_topic=reply_topic,
                          args=kwargs)
            return None, None, None

    def _format_response(self, msg_header, msg_body, status):
        """
        Format a response
        :param msg_header: The header portion of a received request
        :param msg_body: The response body
        :param status: True is this represents a successful response
        :return: a InterContainerMessage message type
        """
        try:
            assert isinstance(msg_header, Header)
            response = InterContainerMessage()
            response_body = InterContainerResponseBody()
            response.header.id = msg_header.id
            response.header.timestamp = int(
                round(time.time() * 1000))
            response.header.type = MessageType.Value("RESPONSE")
            response.header.from_topic = msg_header.to_topic
            response.header.to_topic = msg_header.from_topic
            if msg_body is not None:
                response_body.result.Pack(msg_body)
            response_body.success = status
            response.body.Pack(response_body)
            return response
        except Exception as e:
            log.exception("formatting-response-failed", header=msg_header,
                          body=msg_body, status=status, e=e)
            return None

    def _parse_response(self, msg):
        try:
            message = InterContainerMessage()
            message.ParseFromString(msg)
            resp = InterContainerResponseBody()
            if message.body.Is(InterContainerResponseBody.DESCRIPTOR):
                message.body.Unpack(resp)
            else:
                log.debug("unsupported-msg", msg_type=type(message.body))
                return None
            log.debug("parsed-response", type=message.header.type, from_topic=message.header.from_topic,
                  to_topic=message.header.to_topic)
            return resp
        except Exception as e:
            log.exception("parsing-response-failed", msg=msg, e=e)
            return None

    @inlineCallbacks
    def _process_message(self, m):
        """
        Default internal method invoked for every batch of messages received
        from Kafka.
        """

        def _augment_args_with_FromTopic(args, from_topic):
            arg = Argument(key=ARG_FROM_TOPIC)
            t = StrType(val=from_topic)
            arg.value.Pack(t)
            args.extend([arg])
            return args

        def _toDict(args):
            """
            Convert a repeatable Argument type into a python dictionary
            :param args: Repeatable core_adapter.Argument type
            :return: a python dictionary
            """
            if args is None:
                return None
            result = {}
            for arg in args:
                assert isinstance(arg, Argument)
                result[arg.key] = arg.value
            return result

        current_time = int(time.time() * 1000)
        # log.debug("Got Message", message=m)
        try:
            val = m.value()
            # val = m.message.value
            # print m.topic

            # Go over customized callbacks first
            m_topic = m.topic()
            if m_topic in self.topic_callback_map:
                for c in self.topic_callback_map[m_topic]:
                    yield c(val)

            #  Check whether we need to process request/response scenario
            if m_topic not in self.topic_target_cls_map:
                return

            # Process request/response scenario
            message = InterContainerMessage()
            message.ParseFromString(val)

            if message.header.type == MessageType.Value("REQUEST"):
                # Get the target class for that specific topic
                targetted_topic = self._to_string(message.header.to_topic)
                msg_body = InterContainerRequestBody()
                if message.body.Is(InterContainerRequestBody.DESCRIPTOR):
                    message.body.Unpack(msg_body)
                else:
                    log.debug("unsupported-msg", msg_type=type(message.body))
                    return
                if targetted_topic in self.topic_target_cls_map:
                    # Augment the request arguments with the from_topic
                    augmented_args = _augment_args_with_FromTopic(msg_body.args,
                                                        msg_body.reply_to_topic)
                    if augmented_args:
                        log.debug("message-body-args-present", rpc=msg_body.rpc,
                                  response_required=msg_body.response_required, reply_to_topic=msg_body.reply_to_topic)
                        (status, res) = yield getattr(
                            self.topic_target_cls_map[targetted_topic],
                            self._to_string(msg_body.rpc))(
                            **_toDict(augmented_args))
                    else:
                        log.debug("message-body-args-absent", rpc=msg_body.rpc,
                                  response_required=msg_body.response_required, reply_to_topic=msg_body.reply_to_topic,)
                        (status, res) = yield getattr(
                            self.topic_target_cls_map[targetted_topic],
                            self._to_string(msg_body.rpc))()
                    if msg_body.response_required:
                        response = self._format_response(
                            msg_header=message.header,
                            msg_body=res,
                            status=status,
                        )
                        if response is not None:
                            res_topic = self._to_string(
                                response.header.to_topic)
                            self._send_kafka_message(res_topic, response)

                        log.debug("Response-sent",
                                  to_topic=res_topic)
            elif message.header.type == MessageType.Value("RESPONSE"):
                trns_id = self._to_string(message.header.id)
                if trns_id in self.transaction_id_deferred_map:
                    resp = self._parse_response(val)

                    self.transaction_id_deferred_map[trns_id].callback(resp)
            else:
                log.error("!!INVALID-TRANSACTION-TYPE!!")

        except Exception as e:
            log.exception("Failed-to-process-message", message=m, e=e)

    @inlineCallbacks
    def _send_kafka_message(self, topic, msg):
        try:
            yield self.kafka_proxy.send_message(topic, msg.SerializeToString())
        except Exception as e:
            log.exception("Failed-sending-message", message=msg, e=e)

    @inlineCallbacks
    def send_request(self,
                     rpc,
                     to_topic,
                     reply_topic=None,
                     callback=None,
                     **kwargs):
        """
        Invoked to send a message to a remote container and receive a
        response if required.
        :param rpc: The remote API to invoke
        :param to_topic: Send the message to this kafka topic
        :param reply_topic: If not None then a response is expected on this
        topic.  If set to None then no response is required.
        :param callback: Callback to invoke when a response is received.
        :param kwargs: Key-value pairs representing arguments to pass to the
        rpc remote API.
        :return: Either no response is required, or a response is returned
        via the callback or the response is a tuple of (status, return_cls)
        """
        try:
            # Ensure all strings are not unicode encoded
            rpc = self._to_string(rpc)
            to_topic = self._to_string(to_topic)
            reply_topic = self._to_string(reply_topic)

            request, transaction_id, response_required = \
                self._format_request(
                    rpc=rpc,
                    to_topic=to_topic,
                    reply_topic=reply_topic,
                    **kwargs)

            if request is None:
                return

            # Add the transaction to the transaction map before sending the
            # request.  This will guarantee the eventual response will be
            # processed.
            wait_for_result = None
            if response_required:
                wait_for_result = Deferred()
                self.transaction_id_deferred_map[
                    self._to_string(request.header.id)] = wait_for_result
            log.debug("message-send", transaction_id=transaction_id, to_topic=to_topic,
                      from_topic=reply_topic)
            yield self._send_kafka_message(to_topic, request)
            log.debug("message-sent", transaction_id=transaction_id, to_topic=to_topic,
                      from_topic=reply_topic)

            if response_required:
                res = yield wait_for_result

                # Remove the transaction from the transaction map
                del self.transaction_id_deferred_map[transaction_id]

                if res is not None:
                    if res.success:
                        log.debug("send-message-response", rpc=rpc)
                        if callback:
                            callback((res.success, res.result))
                        else:
                            returnValue((res.success, res.result))
                    else:
                        # this is the case where the core API returns a grpc code.NotFound.  Return or callback
                        # so the caller can act appropriately (i.e add whatever was not found)
                        log.warn("send-message-response-error-result", kafka_request=request, kafka_result=res)
                        if callback:
                            callback((res.success, None))
                        else:
                            returnValue((res.success, None))
                else:
                    raise KafkaMessagingError(error="failed-response-for-request:{}".format(request))

        except Exception as e:
            log.exception("Exception-sending-request", e=e)
            raise KafkaMessagingError(error=e)


# Common method to get the singleton instance of the kafka proxy class
def get_messaging_proxy():
    return IKafkaMessagingProxy._kafka_messaging_instance



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from unittest import TestCase, main
from nose.tools import assert_raises
from pyvoltha.adapters.extensions.omci.me_frame import *
from pyvoltha.adapters.extensions.omci.omci_me import *
from pyvoltha.adapters.extensions.omci.omci import *
import codecs


def hexify(frame):
    """Return a hexadecimal string encoding of input buffer"""
    return codecs.encode(bytes(frame),'hex')

class TestSelectMeFrameGeneration(TestCase):

    def assertGeneratedFrameEquals(self, frame, ref):
        assert isinstance(frame, Packet)
        serialized_hexified_frame = hexify(frame).upper()
        ref = ref.upper()
        if serialized_hexified_frame != ref:
            self.fail('Mismatch:\nReference:\n{}\nGenerated (bad):\n{}'.format(
                ref, serialized_hexified_frame
            ))

    def test_mib_reset_message_serialization(self):
        ref = b'00004F0A000200000000000000000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        frame = OntDataFrame().mib_reset()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_gal_ethernet_profile(self):
        ref = b'0000440A011000010030000000000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        frame = GalEthernetProfileFrame(1, max_gem_payload_size=48).create()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_tcont_1(self):
        ref = b'0000480A010680008000040000000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'

        frame = TcontFrame(0x8000, alloc_id=0x400).set()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_tcont_2(self):
        ref = b'0000480A010680018000040100000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'

        frame = TcontFrame(0x8001, alloc_id=0x401).set()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_8021p_mapper_service_profile(self):
        ref = b'0000440A00828000ffffffffffffffff' \
              b'ffffffffffffffffffff000000000000' \
              b'000000000000000000000028'
        frame = Ieee8021pMapperServiceProfileFrame(0x8000).create()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_mac_bridge_service_profile(self):
        ref = b'0000440A002D02010001008000140002' \
              b'000f0001000000000000000000000000' \
              b'000000000000000000000028'
        data = dict(
            spanning_tree_ind=False,
            learning_ind=True,
            priority=0x8000,
            max_age=20 * 256,
            hello_time=2 * 256,
            forward_delay=15 * 256,
            unknown_mac_address_discard=True
        )
        frame = MacBridgeServiceProfileFrame(0x201, attributes=data).create()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_gem_port_network_ctp(self):
        ref = b'0000440A010C01000400800003010000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'

        data = dict(
            port_id=0x400,
            tcont_pointer=0x8000,
            direction=3,
            traffic_management_pointer_upstream=0x100
        )
        frame = GemPortNetworkCtpFrame(0x100, attributes=data).create()
        self.assertGeneratedFrameEquals(frame, ref)

        # Also test direction as a string parameter
        frame = GemPortNetworkCtpFrame(0x100, port_id=0x400,
                                       tcont_id=0x8000,
                                       direction='bi-directional',
                                       upstream_tm=0x100).create()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_gem_inteworking_tp(self):
        ref = b'0000440A010A80010100058000000000' \
              b'01000000000000000000000000000000' \
              b'000000000000000000000028'
        frame = GemInterworkingTpFrame(0x8001,
                                       gem_port_network_ctp_pointer=0x100,
                                       interworking_option=5,
                                       service_profile_pointer=0x8000,
                                       interworking_tp_pointer=0x0,
                                       gal_profile_pointer=0x1).create()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_8021p_mapper_service_profile(self):
        ref = b'0000480A008280007F80800100000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        ptrs = [0x8001, 0, 0, 0, 0, 0, 0, 0]
        frame = Ieee8021pMapperServiceProfileFrame(0x8000,
                                                   interwork_tp_pointers=ptrs).set()

        self.assertGeneratedFrameEquals(frame, ref)

        ptrs = [0x8001, 0]
        frame = Ieee8021pMapperServiceProfileFrame(0x8000,
                                                   interwork_tp_pointers=ptrs).set()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_mac_bridge_port_configuration_data(self):
        ref = b'0000440A002F21010201020380000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'

        frame = MacBridgePortConfigurationDataFrame(0x2101,
                                                    bridge_id_pointer=0x201,
                                                    port_num=2,
                                                    tp_type=3,
                                                    tp_pointer=0x8000).create()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_vlan_tagging_filter_data(self):
        ref = b'0000440A005421010400000000000000' \
              b'00000000000000000000000000000000' \
              b'100100000000000000000028'
        frame = VlanTaggingFilterDataFrame(0x2101,
                                           vlan_tcis=[0x400],
                                           forward_operation=0x10).create()
        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_extended_vlan_tagging_operation_configuration_data(self):
        ref = b'0000440A00AB02020A04010000000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        data = dict(
            association_type=10,
            associated_me_pointer=0x401
        )
        frame = \
            ExtendedVlanTaggingOperationConfigurationDataFrame(0x202,
                                                               attributes=data)\
                .create()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_extended_vlan_tagging_operation_configuration_data(self):
        ref = b'0000480A00AB02023800810081000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        data = dict(
            input_tpid=0x8100,
            output_tpid=0x8100,
            downstream_mode=0,  # inverse of upstream
        )
        frame = \
            ExtendedVlanTaggingOperationConfigurationDataFrame(0x202,
                                                               attributes=data)\
                .set()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_extended_vlan_tagging_1(self):
        ref = b'0000480A00AB02020400f00000008200' \
              b'5000402f000000082004000000000000' \
              b'000000000000000000000028'
        data = dict(
            received_frame_vlan_tagging_operation_table=\
                VlanTaggingOperation(
                    filter_outer_priority=15,
                    filter_inner_priority=8,
                    filter_inner_vid=1024,
                    filter_inner_tpid_de=5,
                    filter_ether_type=0,
                    treatment_tags_to_remove=1,
                    pad3=2,
                    treatment_outer_priority=15,
                    treatment_inner_priority=8,
                    treatment_inner_vid=1024,
                    treatment_inner_tpid_de=4
                )
        )
        frame = \
            ExtendedVlanTaggingOperationConfigurationDataFrame(0x202,
                                                               attributes=data)\
                .set()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_extended_vlan_tagging_2(self):
        ref = b'0000480A00AB02020400F00000008200' \
              b'd000402f00000008200c000000000000' \
              b'000000000000000000000028'
        data = dict(
            received_frame_vlan_tagging_operation_table=
                VlanTaggingOperation(
                    filter_outer_priority=15,
                    filter_inner_priority=8,
                    filter_inner_vid=1025,
                    filter_inner_tpid_de=5,
                    filter_ether_type=0,
                    treatment_tags_to_remove=1,
                    pad3=2,
                    treatment_outer_priority=15,
                    treatment_inner_priority=8,
                    treatment_inner_vid=1025,
                    treatment_inner_tpid_de=4
                )
        )

        frame = \
            ExtendedVlanTaggingOperationConfigurationDataFrame(0x202,
                                                               attributes=data)\
                .set()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_create_mac_bridge_port_configuration_data2(self):
        ref = b'0000440A002F02010201010b04010000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        data = dict(
            bridge_id_pointer=0x201,
            encapsulation_methods=0,
            port_num=1,
            port_priority=0,
            port_path_cost=0,
            port_spanning_tree_in=0,
            lan_fcs_ind=0,
            tp_type=11,
            tp_pointer=0x401,
            mac_learning_depth=0
        )
        frame = MacBridgePortConfigurationDataFrame(0x201,
                                                    attributes=data).create()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_set_pptp_ethernet_uni_frame(self):
        ref = b'0000480A000B020109000005EE000000' \
              b'00000000000000000000000000000000' \
              b'000000000000000000000028'
        data = dict(
            administrative_state=0,  # 0 - Unlock
            max_frame_size=1518      # two-octet field
        )
        frame = PptpEthernetUniFrame(0x201,
                                     attributes=data).set()

        self.assertGeneratedFrameEquals(frame, ref)

    def test_constraint_errors(self):
        self.assertTrue(True)  # TODO Also test some attribute constraint failures

    def test_mib_upload_next(self):
        # Test for VOL-649 error. SCAPY was only originally coded for a 'get'
        # action (8-bit MIB Data Sync value) but MIB Upload Next commands have
        # a 16-bit field.
        #
        # 255 and less always worked
        OntDataFrame(sequence_number=0).mib_upload_next()
        OntDataFrame(sequence_number=255).mib_upload_next()
        # But not 256+
        OntDataFrame(sequence_number=256).mib_upload_next()
        OntDataFrame(sequence_number=1000).mib_upload_next()
        OntDataFrame(sequence_number=0xFFFE).mib_upload_next()

        # Also test the optional arguments for the other actions
        OntDataFrame().get()
        OntDataFrame(mib_data_sync=4).set()
        OntDataFrame().mib_reset()
        OntDataFrame().mib_upload()
        # OntDataFrame(ignore_arc=True).get_all_alarms()        Not yet coded
        # OntDataFrame(ignore_arc=False).get_all_alarms()       Not yet coded

        # Range/type checks
        assert_raises(ValueError, OntDataFrame, mib_data_sync=-1)
        assert_raises(ValueError, OntDataFrame, mib_data_sync=256)
        assert_raises(ValueError, OntDataFrame, sequence_number=-1)
        assert_raises(ValueError, OntDataFrame, sequence_number=0x10000)
        assert_raises(TypeError, OntDataFrame, ignore_arc=123)


if __name__ == '__main__':
    main()




# Copyright 2017-present Adtran, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from voltha_protos.events_pb2 import EventCategory, EventSubCategory, EventType
from pyvoltha.adapters.extensions.events.adapter_events import DeviceEventBase


class OnuDisabledEvent(DeviceEventBase):
    def __init__(self, event_mgr, device_id, pon_id, onu_serial_number,
                 reg_id, olt_serial_number, raised_ts, ipv4_address=None,
                 onu_id=None):
        super(OnuDisabledEvent, self).__init__(event_mgr, raised_ts, object_type='ONU',
                                             event='ONU_DISABLED',
                                             resource_id=pon_id,
                                             category=EventCategory.COMMUNICATION,
                                             sub_category=EventSubCategory.PON,
                                             )

        self._pon_id = pon_id
        self._onu_id = onu_id
        self._onu_serial_number = onu_serial_number
        self._device_id = device_id
        self._olt_serial_number = olt_serial_number
        self._host = ipv4_address
        self._reg_id = reg_id

    def get_context_data(self):
        data = {
            'pon-id': self._pon_id,
            'onu-id': self._onu_id,
            'serial-number': self._onu_serial_number,
            'olt_serial_number': self._olt_serial_number,
            'device_id': self._device_id,
            'registration_id': self._reg_id
        }
        if self._host is not None:
            data['host'] = self._host

        return data



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# import binascii
from __future__ import absolute_import
import structlog
import six
# from twisted.internet.defer import Deferred
# from voltha.core.config.config_root import ConfigRoot
# from pyvoltha.protos.voltha_pb2 import VolthaInstance
# from pyvoltha.adapters.extensions.omci.omci_frame import OmciFrame


class MockProxyAddress(object):
    def __init__(self, device_id, pon_id, onu_id):
        self.device_id = device_id  # Device ID of proxy (OLT)
        self.onu_id = onu_id
        self.onu_session_id = onu_id

        self.channel_group_id = pon_id  # close enough for mock
        self.channel_id = pon_id
        self.channel_termination = pon_id


class MockDevice(object):
    def __init__(self, device_id, proxy_address=None, serial_number=None):
        from pyvoltha.adapters.extensions.omci.omci_entities import entity_id_to_class_map
        self.id = device_id
        self.parent_id = None
        self.proxy_address = proxy_address
        self.serial_number = serial_number
        self.me_map = entity_id_to_class_map


class MockCore(object):
    def __init__(self):
        self.root = None   # ConfigRoot(VolthaInstance())
        if self.root is None:
            from nose import SkipTest
            msg = "TODO: Unit tests involving the kv-store are not yet supported by pyvoltha"
            raise SkipTest(msg)

    def get_proxy(self, path):
        return self.root.get_proxy(path)


class MockAdapterAgent(object):
    """
    Minimal class to handle adapter-agent needs in OpenOMCI. It can be
    used by a mock OLT or ONU.

    So that we do not have to duplicate the IAdapter functionality, just
    the handler, the OLT and ONU handlers are derived from a mock Device
    base class so that we can access the _devices map and get either a
    Device to play with (like the real thing) or the associated handler
    """
    def __init__(self, d=None):
        self.log = structlog.get_logger() 
        self._devices = dict()      # device-id -> mock device
        self.core = MockCore()
        self.deferred = d
        self.timeout_the_message = False

    @property
    def send_omci_defer(self):
        return self.deferred
        
    @send_omci_defer.setter
    def send_omci_defer(self, value):
        self.deferred = value

    @property
    def name(self):
        return "cig_mock_ont"
    
    def tearDown(self):
        """Test case cleanup"""
        for device in six.itervalues(self._devices):
            device.tearDown()
        self._devices.clear()

    def add_device(self, device):
        self._devices[device.id] = device

    def add_child_device(self, parent_device, child_device):
        # Set parent
        child_device.parent_id = parent_device.id

        # Add ONU serial number if PON and ONU enabled

        if (child_device.enabled and
                child_device.serial_number is not None and
                child_device.proxy_address.channel_id in parent_device.enabled_pons):
            parent_device.activated_onus.add(child_device.serial_number)

        self.add_device(child_device)

    def get_device(self, device_id):
        return self._devices[device_id]

    def get_child_device(self, parent_device_id, **kwargs):
        onu_id = kwargs.pop('onu_id', None)
        pon_id = kwargs.pop('pon_id', None)
        if onu_id is None and pon_id is None:
            return None

        # Get all child devices with the same parent ID
        children_ids = set(d.id for d in six.itervalues(self._devices)
                           if d.parent_id == parent_device_id)

        # Loop through all the child devices with this parent ID
        for child_id in children_ids:
            device = self.get_device(child_id)

            # Does this child device match the passed in ONU ID?
            found_onu_id = False
            if onu_id is not None:
                if device.proxy_address.onu_id == onu_id:
                    found_onu_id = True

            # Does this child device match the passed in SERIAL NUMBER?
            found_pon_id = False
            if pon_id is not None:
                if device.proxy_address.channel_id == pon_id:
                    found_pon_id = True
            # Match ONU ID and PON ID
            if onu_id is not None and pon_id is not None:
                found = found_onu_id & found_pon_id
            # Otherwise ONU ID or PON ID
            else:
                found = found_onu_id | found_pon_id

            # Return the matched child device
            if found:
                return device

        return None

    def send_proxied_message(self, proxy_address, msg):
        # Look up ONU handler and forward the message
        self.log.debug("--> send_proxied_message", message=msg)
        
        # if proxy_address is None:
        if self.deferred is not None and not self.timeout_the_message:
            self.deferred.callback(msg)
        #     return None

        # olt_handler = self.get_device(proxy_address.device_id)

        # if olt_handler is not None:
        #    olt_handler.send_proxied_message(proxy_address, msg)

    def receive_proxied_message(self, proxy_address, msg):
        # Look up ONU handler and forward the message

        onu_handler = self.get_child_device(proxy_address.device_id,
                                            onu_id=proxy_address.onu_id,
                                            pon_id=proxy_address.channel_id)
        if onu_handler is not None:
            onu_handler.receive_proxied_message(proxy_address, msg)



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from enum import Enum, IntEnum

class OmciUninitializedFieldError(Exception):
    pass


class OmciInvalidTypeError(Exception):
    pass

def bitpos_from_mask(mask, lsb_pos=0, increment=1):
    """
    Turn a decimal value (bitmask) into a list of indices where each
    index value corresponds to the bit position of a bit that was set (1)
    in the mask. What numbers are assigned to the bit positions is controlled
    by lsb_pos and increment, as explained below.
    :param mask: a decimal value used as a bit mask
    :param lsb_pos: The decimal value associated with the LSB bit
    :param increment: If this is +i, then the bit next to LSB will take
    the decimal value of lsb_pos + i.
    :return: List of bit positions where the bit was set in mask
    """
    out = []
    while mask:
        if mask & 0x01:
            out.append(lsb_pos)
        lsb_pos += increment
        mask >>= 1
    return sorted(out)


class AttributeAccess(Enum):
    Readable = 1
    R = 1
    Writable = 2
    W = 2
    SetByCreate = 3
    SBC = 3


OmciNullPointer = 0xffff
OmciSectionDataSize = 31

class EntityOperations(Enum):
    # keep these numbers match msg_type field per OMCI spec
    Create = 4
    CreateComplete = 5
    Delete = 6
    Set = 8
    Get = 9
    GetComplete = 10
    GetAllAlarms = 11
    GetAllAlarmsNext = 12
    MibUpload = 13
    MibUploadNext = 14
    MibReset = 15
    AlarmNotification = 16
    AttributeValueChange = 17
    Test = 18
    StartSoftwareDownload = 19
    DownloadSection = 20
    EndSoftwareDownload = 21
    ActivateSoftware = 22
    CommitSoftware = 23
    SynchronizeTime = 24
    Reboot = 25
    GetNext = 26
    TestResult = 27
    GetCurrentData = 28
    SetTable = 29       # Defined in Extended Message Set Only


class ReasonCodes(IntEnum):
    # OMCI Result and reason codes
    Success = 0,            # Command processed successfully
    ProcessingError = 1,    # Command processing error
    NotSupported = 2,       # Command not supported
    ParameterError = 3,     # Parameter error
    UnknownEntity = 4,      # Unknown managed entity
    UnknownInstance = 5,    # Unknown managed entity instance
    DeviceBusy = 6,         # Device busy
    InstanceExists = 7,     # Instance Exists
    AttributeFailure = 9,   # Attribute(s) failed or unknown

    OperationCancelled = 255 # Proprietary defined for internal use
    
